from sqlalchemy.exc import IntegrityError

from asc.core import log
from asc.core.ABC.Control import ControlTool
from asc.tools.marketplace.mappings import PaymentMethod
from asc.tools.marketplace.mappings import PaymentGateway
from asc.tools.marketplace.mercadopago import data as mpg_data
from asc.tools.marketplace.mercadopago.api import APIMercadoPago
from asc.core.filters.filter_crypme import Crypme
from asc.models.marketplace import User as MarketPlaceUser, UserCard
from asc.models.mercadopago import Payment, PaymentMethod
from asc.models.mercadopago import User as MercadoPagoUser, Card
from asc.exceptions import marketplace as exceptions_marketplace
from asc.exceptions import mercadopago as exceptions_mercadopago

cryp = Crypme()


class ControlToolHandler(ControlTool):

    __payment_gateway:PaymentGateway = None

    def __init__(self):
        super(ControlToolHandler, self).__init__()

        self.__MP = APIMercadoPago()

    def __authorize_card(self, id_user:int, id_user_card:int):
        """
        Atualiza token do cartão para utiliza-lo em compras.
        :param id_user:
        :param id_user_card:
        :return:
        """
        card_token = self.generate_card_token(
            id_user_card=id_user_card
        )
        card_customer = self.apply_card_token(
            id_user=id_user,
            id_user_card=id_user_card,
            card_token=card_token['id']
        )
        return card_customer

    def __mount_payment_pix(self):
        return mpg_data.PaymentPix()

    def __mount_payment_card(self, id_user_card:int, installments:int=1, **card_auth):
        customer_card = self.session \
            .query(
                Card.secret,
                Card.id_payment_method) \
            .where(
                Card.id_user_card == id_user_card) \
            .join(
                PaymentMethod,
                PaymentMethod.id == Card.id_payment_method) \
            .first()

        return mpg_data.PaymentCard(
            token=customer_card.secret,
            issuer_id=card_auth['issuer']['id'],
            payment_method_id=customer_card.id_payment_method,
            installments=installments
        )

    def __query_customer(self, id_user:int, *select):
        if not select:
            select = MarketPlaceUser

        user = self.session \
            .query(*select) \
            .join(MercadoPagoUser,
                  MercadoPagoUser.id_user
                  == MarketPlaceUser.id_user, isouter=True) \
            .where(MarketPlaceUser.id_user == id_user) \
            .first()

        if not user:
            raise exceptions_marketplace.UserDoesNotExists(id_user)

        return user

    def _commit_payment(self, id_payment:int, http_response:dict):
        payment = Payment()
        payment.id_payment = id_payment
        payment.http_response = http_response
        payment.status = http_response.get('status')
        payment.taxes_amount = http_response.get('taxes_amount')
        payment.transaction_amount = http_response.get('transaction_amount')
        payment.id_mercadopago = http_response.get('id')
        self.session.add(payment)
        self.commit()
        return True

    def _commit_payment_postback(self, id_payment:int, http_postback:dict):
        payment = self.session \
            .query(Payment) \
            .where(Payment.id_payment == id_payment) \
            .first()

        if not payment:
            return False

        payment.http_postback = http_postback
        payment.status = http_postback.get('status')
        self.commit()
        return True

    def _calc_installments(self, method:PaymentMethod, amount:int):
        pass

    def save_payment_methods(self) -> bool:
        """ Busca e registra métodos de pagamentos aceitos.
        :rtype: object
        :return: bool
        """
        payment_methods = self.__MP.GET_payment_methods()
        # Encontra código em cache
        for method in payment_methods:
            model = PaymentMethod()
            model.id = method.get('id')
            model.name = method.get('name')
            model.id_payment_type = method.get('payment_type_id')
            model.active = bool(method.get('status') == 'active')
            self.session.add(model)

        self.commit()
        return True

    def get_client_id(self, id_user:int) -> int:
        """ Registra ou obtém mercadopago_user.id_customer.
        :return: ID do cliente registrado no mercado pago.
        """
        user = self.\
            __query_customer(
            id_user,
            MarketPlaceUser.name,
            MarketPlaceUser.mail,
            MercadoPagoUser.id_client
        )

        customer_id = user.id_client
        customer_name = user.name.split(' ')

        if not customer_id:
            try:
                customer_data = self.__MP.POST_customer(
                    email=user.mail,
                    first_name=customer_name[0],
                    last_name=customer_name[-1]
                )

                customer = MercadoPagoUser()
                customer.id = id_user
                customer.id_customer = customer_data['id']
                customer.id_client = customer_data['client_id']
                customer.id_user = customer_data['user_id']
                self.session.add(customer)
                customer_id = customer.id_customer

            except (exceptions_mercadopago \
                    .CustomerAlreadyExistsException,
                    IntegrityError) as e:
                log.error(e)
                customer = self.__MP.GET_customer_search(user.mail)
                customer_id = customer['results'][0]['id']

        return customer_id

    def generate_card_token(self, id_user_card:int):
        """
        Registra cartão na mercadopago.
        :param id_user_card:
        :return:
        """

        card = self.session \
            .query(
                UserCard.nr_card,
                UserCard.nr_cvv,
                UserCard.dt_add,
                UserCard.dt_exp,
                MarketPlaceUser.name,
                MarketPlaceUser.udoc) \
            .join(MarketPlaceUser,
                  MarketPlaceUser.id_user
                  == UserCard.id_user) \
            .where(UserCard.id_user_card
                   ==id_user_card) \
            .first()

        cryp.getGrid(card.dt_add)
        card_holder = mpg_data.CardHolder(
            name=card.name,
            identification=dict(CPF=card.udoc)
        )
        card = mpg_data.Card(
            card_number=cryp.btoa(card.nr_card),
            security_code=card.nr_cvv,
            expiration_month=card.dt_exp[:2],
            expiration_year=f'20{card.dt_exp[2:]}',
            cardholder=card_holder
        )

        POST_card = self.__MP.POST_card(card)

        if not 'id' in POST_card:
            raise Exception('ID não encontrado em POST_card')

        return POST_card

    def apply_card_token(self, card_token:str, id_user, id_user_card):
        """
        Renova ou cria card_token do usuário.
        :param card_token:
        :param id_user: user.id_user
        :param id_user: user_card.id_user_card
        :return:
        """

        customer_card = self.__MP.POST_customer_card(
            id_customer=self.get_client_id(id_user),
            token=card_token
        )

        mercadopago_card = self.session \
            .query(Card) \
            .where(Card.id_user_card == id_user_card) \
            .scalar()

        if not mercadopago_card:
            mercadopago_card = Card()

        mercadopago_card.secret = card_token
        mercadopago_card.id_payment_method = customer_card \
            .get('payment_method')['id']

        if not mercadopago_card.id_user_card:
            # Quando ainda não existir relação, INSERT.
            mercadopago_card.id_user_card = id_user_card
            self.session.add(mercadopago_card)

        return customer_card

    def get_payment(self, id_payment:int):
        return self.__MP.GET_payment(id_payment)

    def send_payment(self, id_user:int, amount:int, **props):
        """
        Solicita pagamento com a mercadopago.
        :param id_user: usuário responsável.
        :param amount: valor da transação.
        :param id_user_card:
        :return:
        """
        id_user_card:int = props.get('id_user_card')

        customer_user = self. \
            __query_customer(
                id_user,
                MarketPlaceUser.name,
                MarketPlaceUser.udoc,
                MarketPlaceUser.utel,
                MarketPlaceUser.mail,
                MercadoPagoUser.id_customer
        )

        customer_user = customer_user._asdict()

        if id_user_card:
            card_authorized = self.__authorize_card(
                id_user=id_user,
                id_user_card=id_user_card
            )
            log.info(card_authorized)
            payment = self.__mount_payment_card(
                id_user_card=id_user_card,
                installments=props.get('installments',1),
                **card_authorized
            )
        else:
            payment = self.__mount_payment_pix()

        payment.update(props)
        payment.transaction_amount = amount
        payment.payer.email = customer_user['mail']
        payment.payer.mount_name(customer_user['name'])
        payment.payer.mount_identification(customer_user['udoc'])

        return self.__MP.POST_payment(
            payment
        )